==> parampl-0.0.7-py3-none-any.whl: parampl/core.py <==
import numpy as np
from matplotlib.axes import Axes

from parampl.statics import (split_into_paragraphs, parse_avoid,
                             avoid_specification, avoid_single_specification, get_aspect)


__all__ = ['ParaMPL', 'avoid_specification', 'avoid_single_specification']


class _line_position:
    def __init__(self,
                 xy,
                 width, height,
                 rotation, spacing,
                 ha, justify,
                 y_to_x_ratio=1.0,
                 xy_at_top=True):
        self.x_orig, self.y_orig = xy
        self.width = width
        self.height = height
        self.rotation = rotation

        if ha == 'right':
            self.x_orig -= width
        elif ha == 'center':
            self.x_orig -= width / 2.0
        elif ha != 'left':
            raise ValueError(f"invalid ha '{ha}'. Must be 'right', 'left', or 'center'")

        if xy_at_top:
            self.y_orig -= height * np.cos(rotation * np.pi / 180)  # top alignment
            self.x_orig -= height * np.sin(rotation * np.pi / 180)  # top alignment

        self.delta_x = (1 + spacing) * height * np.sin(rotation * np.pi / 180) * y_to_x_ratio
        self.delta_y = - (1 + spacing) * height * np.cos(rotation * np.pi / 180)

        self.borders = [(None, self.x_orig, width)]

        self.limit = None
        self.x = self.x_orig
        self.y = self.y_orig
        self.width_line = self.width

        self.justify_mult = (justify == 'right') + 0.5 * (justify == 'center')
        if justify not in ['right', 'center', 'left', 'full']:
            raise ValueError(f'Unrecognized justify {justify}')

    def check_next_border(self, force=False):
        if force:
            self.limit, self.x, self.width_line = self.borders.pop(0)
        while self.limit is not None and self.y < self.limit:
            self.limit, self.x, self.width_line = self.borders.pop(0)

    def add_avoids(self, avoid_left_of, avoid_right_of):
        if avoid_left_of is not None or avoid_right_of is not None:
            self.borders = parse_avoid(self.borders, avoid_left_of, avoid_right_of, self.height)

    def offset(self,
               offset: float = 0,
               justified_length: float = 0):
        total_offset = self.justify_mult * (self.width_line - justified_length) + offset
        return (self.x + total_offset * np.cos(self.rotation * np.pi / 180),
                self.y + total_offset * np.sin(self.rotation * np.pi / 180))

    def next_line(self):
        self.x += self.delta_x
        self.y += self.delta_y

        self.check_next_border()

    def total_height(self):
        return ((self.y_orig - self.y) * np.cos(self.rotation * np.pi / 180) +
                self.width * np.sin(self.rotation*np.pi / 180))


class ParaMPL:
    def __init__(self,
                 axes: Axes,
                 width: float = 1.0,
                 spacing: float = 0.5,
                 fontsize: float = 10,
                 justify: str = "left",
                 fontname: str | None = None,
                 family: str | None = None,
                 color: None | str | tuple[float, float, float] = None,
                 transform: str = 'data',
                 ):
        """

        :param axes:
          matplotlib.axes.Axes in which to put the paragraphs
        :param spacing:
          default spacing
        :param width:
           default width
        :param fontsize:
           default fontsize
        :param color:
          default color
        :param transform:
          transform in which the coordinates are given. Currently supported: 'data'
        """
        self.width = width
        self.spacing = spacing
        self.axes = axes
        self.fontsize = fontsize
        self.color = color
        self.family = family
        self.fontname = fontname
        self.justify = justify

        self._renderer = axes.get_figure().canvas.get_renderer()
        if transform == 'data':
            self._transform = axes.transData.inverted()
        else:
            raise NotImplementedError("only 'data' transform is supported for now")

        self.widths: dict[tuple, dict[str, float]] = {}
        self.heights: dict[tuple, float] = {}

    def write(self,
              text: str,
              xy: tuple[float, float],
              width: float | None = None,
              spacing: float | None = None,
              fontsize: float | None = None,
              color: str | None = None,
              fontname: str | None = None,
              family: str | None = None,
              rotation: float = 0,
              justify: str | None = None,
              ha: str = 'left',
              va: str = 'top',
              avoid_left_of: avoid_specification = None,
              avoid_right_of: avoid_specification = None,
              collapse_whites: bool = True,
              paragraph_per_line: bool = False,
              ):
        """
Write text into a paragraph

        :param text:
          text to write
        :param xy:
           xy to place the paragraph
        :param width:
          use this width instead of the initialized one
        :param paragraph_per_line:
          if true, each new line is considered a new paragraph
        :param family:
          family of the font
        :param fontname:
          specific fontname, if not specified then use family
        :param rotation:
           anticlockwise rotation
        :param collapse_whites:
          whether multiple side-by-side withes should be considered as one
        :param color:
          color of text
        :param avoid_left_of:
          tuple (x_lim, (y1, y2)). Avoid space left of x_lim between y1 and y2
        :param avoid_right_of:
          tuple (x_lim, (y1, y2)). Avoid space right of x_lim between y1 and y2
        :param va:
          Paragraph vertical alignment
        :param ha:
          Paragraph horizontal alignment
        :param justify:
          Line's justification
        :param spacing:
          use this spacing instead of the initialized one
        :param fontsize:
          use this fontsize instead of the initialized one
        """

        if width is None:
            width = self.width
        if justify is None:
            justify = self.justify
        if spacing is None:
            spacing = self.spacing
        if fontsize is None:
            fontsize = self.fontsize
        if color is None:
            color = self.color
        if family is None:
            family = self.family
        if fontname is None:
            fontname_dict = {}
        else:
            fontname_dict = {'fontname': self.fontname}

        ax = self.axes

        def write_line(left, bottom, text_in_line, zorder=3):

            ax.text(left, bottom, text_in_line,
                    fontsize=fontsize, color=color, rotation=rotation,
                    family=family, zorder=zorder, **fontname_dict)

        old_artists = list(ax.texts)

        if ax.get_ylim()[1] < ax.get_ylim()[0] or ax.get_xlim()[1] < ax.get_xlim()[0]:
            raise NotImplementedError("paraMPL.write() is only available for plots with increasing x- and y-axis")

        if va != 'top' and (avoid_left_of is not None or avoid_right_of is not None):
            raise ValueError("if using avoid areas, then va='top' must be used")

        widths, height, combined_hash = self._get_widths_height(fontsize, family, fontname,
                                                                words=text.split())
        space_width = widths[' ']

        lp = _line_position(xy, width, height, rotation, spacing, ha, justify,
                            y_to_x_ratio=get_aspect(ax))
        lp.add_avoids(avoid_left_of, avoid_right_of)
        lp.check_next_border(force=True)

        paragraphs = split_into_paragraphs(text,
                                           collapse_whites=collapse_whites,
                                           paragraph_per_line=paragraph_per_line,
                                           )

        for paragraph in paragraphs:
            words = []
            length = 0

            if justify == 'full':
                for word in paragraph.split(' '):
                    if length + widths[word] > lp.width_line:
                        if len(words) > 1:
                            extra_spacing = (lp.width_line - length + space_width) / (len(words) - 1)
                        else:
                            extra_spacing = 0

                        offset = 0
                        for old_word in words:
                            write_line(*lp.offset(offset=offset),
                                       old_word)
                            offset += extra_spacing + space_width + widths[old_word]

                        lp.next_line()
                        length = 0
                        words = []

                    length += widths[word] + space_width
                    words.append(word)

                write_line(*lp.offset(), ' '.join(words))
                lp.next_line()

            else:
                for word in paragraph.split(' '):
                    if length + widths[word] > lp.width_line:
                        write_line(*lp.offset(justified_length=length - space_width),
                                   ' '.join(words))
                        lp.next_line()
                        length, words = 0, []

                    length += widths[word] + space_width
                    words.append(word)

                write_line(*lp.offset(justified_length=length - space_width),
                           ' '.join(words))
                lp.next_line()

        total_height = lp.total_height()
        lowest = min(lp.y,
                     lp.y + self.width * np.sin(lp.rotation*np.pi/180),
                     lp.y_orig + lp.delta_y,
                     lp.y_orig + lp.delta_y + self.width * np.sin(lp.rotation*np.pi/180),
                     )
        delta = lp.y_orig - lowest

        if va == 'top':
            for artist in ax.texts:
                if artist not in old_artists:
                    artist.set_y(artist.get_position()[1] + delta - total_height)

        elif va == 'bottom':
            for artist in ax.texts:
                if artist not in old_artists:
                    artist.set_y(artist.get_position()[1] + delta)

        elif va == 'center':
            for artist in ax.texts:
                if artist not in old_artists:
                    artist.set_y(artist.get_position()[1] + delta - total_height / 2)

        else:
            raise ValueError(f"invalid va '{va}'. Must be 'top', 'bottom', or 'center'")

    def _get_widths_height(self, fontsize, family, fontname,
                           words: list[str] = None,
                           ):
        text_artist = self.axes.text(0, 0, ' ',
                                     fontsize=fontsize, fontname=fontname, family=family)
        combined_hash = (fontsize, family, fontname)

        if combined_hash not in self.widths:
            text_artist.set_text(' ')
            widths: dict[str, float] = {' ': self._transformed_artist_extent(text_artist).width,
                                        '': 0,
                                        }

            text_artist.set_text('L')
            height = self._transformed_artist_extent(text_artist).height

            self.widths[combined_hash] = widths
            self.heights[combined_hash] = height
        else:
            widths = self.widths[combined_hash]

        if words is not None:
            for word in words:
                if word not in widths:
                    text_artist.set_text(word)
                    widths[word] = self._transformed_artist_extent(text_artist).width

        text_artist.remove()

        return self.widths[combined_hash], self.heights[combined_hash], combined_hash

    def _transformed_artist_extent(self, artist):
        extent = artist.get_window_extent(renderer=self._renderer)
        return extent.transformed(self._transform)

==> parampl-0.1.0-py3-none-any.whl: parampl/core.py <==
import matplotlib.artist
import numpy as np
from matplotlib.axes import Axes

from parampl.statics import (split_into_paragraphs, parse_avoid,
                             avoid_specification, avoid_single_specification, get_aspect)

rectangle_specification = tuple[float, float, float, float]  # left, right, bottom, top

__all__ = ['ParaMPL', 'avoid_specification', 'avoid_single_specification']


class _line_position:
    def __init__(self,
                 xy,
                 width, height,
                 rotation, spacing,
                 ha, justify,
                 y_to_x_ratio=1.0,
                 xy_at_top=True):
        self.x_orig, self.y_orig = xy
        self.width = width
        self.height = height
        self.rotation = rotation

        if ha == 'right':
            self.x_orig -= width
        elif ha == 'center':
            self.x_orig -= width / 2.0
        elif ha != 'left':
            raise ValueError(f"invalid ha '{ha}'. Must be 'right', 'left', or 'center'")

        if xy_at_top:
            self.y_orig -= height * np.cos(rotation * np.pi / 180)  # top alignment
            self.x_orig -= height * np.sin(rotation * np.pi / 180)  # top alignment

        self.delta_x = (1 + spacing) * height * np.sin(rotation * np.pi / 180) * y_to_x_ratio
        self.delta_y = - (1 + spacing) * height * np.cos(rotation * np.pi / 180)

        self.borders = [(None, self.x_orig, width)]
        self.limit: float | None = None
        self.x = self.x_orig
        self.width_line = self.width

        self.y = self.y_orig

        self.justify_mult = (justify == 'right') + 0.5 * (justify == 'center')
        if justify not in ['right', 'center', 'left', 'full']:
            raise ValueError(f'Unrecognized justify {justify}')

    def check_next_border(self, force=False):
        if force:
            self.limit, self.x, self.width_line = self.borders.pop(0)
        while self.limit is not None and self.y < self.limit:
            self.limit, self.x, self.width_line = self.borders.pop(0)

    def add_rectangles(self, rectangles):
        x = self.x_orig
        w = self.width

        avoid_left = []
        avoid_right = []
        for left, right, bottom, top in rectangles:
            left_space = left - x
            right_space = x + w - right
            if left_space < right_space:
                avoid_left.append((right, (bottom, top)))
            else:
                avoid_right.append((left, (bottom, top)))

        self.add_avoids(avoid_left, avoid_right)

    def add_avoids(self, avoid_left_of, avoid_right_of):
        if avoid_left_of is not None or avoid_right_of is not None:
            self.borders = parse_avoid(self.borders, avoid_left_of, avoid_right_of, self.height)

            self.check_next_border(force=True)

    def offset(self,
               offset: float = 0,
               justified_length: float = 0):
        total_offset = self.justify_mult * (self.width_line - justified_length) + offset
        return (self.x + total_offset * np.cos(self.rotation * np.pi / 180),
                self.y + total_offset * np.sin(self.rotation * np.pi / 180))

    def next_line(self):
        self.x += self.delta_x
        self.y += self.delta_y

        self.check_next_border()

    def total_height(self):
        return ((self.y_orig - self.y) * np.cos(self.rotation * np.pi / 180) +
                self.width * np.sin(self.rotation*np.pi / 180))

    def y_to_bottom_offset(self):
        lowest = min(self.y,
                     self.y + self.width * np.sin(self.rotation*np.pi/180),
                     self.y_orig + self.delta_y,
                     self.y_orig + self.delta_y + self.width * np.sin(self.rotation*np.pi/180),
                     )
        return self.y_orig - lowest


class ParaMPL:
    def __init__(self,
                 axes: Axes,
                 transform: str = 'data',

                 width: float = 1.0,
                 spacing: float = 0.5,

                 fontname: str | None = None,
                 fontsize: float = None,
                 family: str | None = None,
                 weight: str | None = None,
                 style: str | None = None,

                 color: None | str | tuple[float, float, float] = None,
                 rotation: float = 0.0,
                 justify: str = "left",
                 ):
        """

        :param axes:
          matplotlib.axes.Axes in which to put the paragraphs
        :param spacing:
          default spacing
        :param width:
           default width
        :param family:
           default font name
        :param family:
           default font family, uses matplotlib's value at initialization if not specified
        :param fontsize:
           default fontsize, uses matplotlib's value at initialization if not specified
        :param weight:
           default weight, uses matplotlib's value at initialization if not specified
        :param style:
           default style, uses matplotlib's value at initialization if not specified
        :param color:
          default color
        :param transform:
          transform in which the coordinates are given. Currently supported: 'data'
        """
        self._width = width
        self._spacing = spacing
        self._axes = axes
        self._color = color
        self._fontname = fontname
        self._justify = justify
        self._rotation = rotation

        if family is None:
            self._family = matplotlib.rcParams['font.family'][0]
        else:
            self._family = family
        if fontsize is None:
            self._fontsize = matplotlib.rcParams['font.size']
        else:
            self._fontsize = fontsize
        if weight is None:
            self._weight = matplotlib.rcParams['font.weight']
        else:
            self._weight = weight
        if style is None:
            self._style = matplotlib.rcParams['font.style']
        else:
            self._style = style

        self._renderer = axes.get_figure().canvas.get_renderer()
        if transform == 'data':
            self._transform = axes.transData.inverted()
        else:
            raise NotImplementedError("only 'data' transform is supported for now")

        self._widths: dict[tuple, dict[str, float]] = {}
        self._heights: dict[tuple, float] = {}
        self._rectangles: list[rectangle_specification] = []

    def avoid_rectangle(self,
                        left_right_bottom_top: rectangle_specification,
                        ):
        """
        Add rectangles to avoid whenever ha='left', va='top', rotation=0 on write()

        :param left_right_bottom_top:
          specify rectangle limits to avoid
        """
        self._rectangles.append(left_right_bottom_top)

    def reset_rectangles(self):
        """Reset avoidance rectangles"""
        self._rectangles = []

    def write(self,
              text: str,
              xy: tuple[float, float],

              width: float | None = None,
              spacing: float | None = None,

              fontname: str | None = None,
              fontsize: float | None = None,
              family: str | None = None,
              weight: str | None = None,
              style: str | None = None,

              color: str | None = None,
              rotation: float | None = None,
              justify: str | None = None,

              ha: str = 'left',
              va: str = 'top',

              avoid_left_of: avoid_specification = None,
              avoid_right_of: avoid_specification = None,
              avoid_rectangles: bool = True,

              collapse_whites: bool = True,
              paragraph_per_line: bool = False,
              ) -> list[matplotlib.artist.Artist]:
        """
Write text into a paragraph

        :rtype:
          return list of artists with text
        :param text:
          text to write
        :param xy:
           xy to place the paragraph

        :param width:
          use this width instead of the initialized one
        :param spacing:
          use this spacing instead of the initialized one

        :param fontsize:
          use this fontsize instead of the initialized one
        :param fontname:
          specific fontname, if not specified then use family
        :param family:
          family of the font
        :param weight:
           font weight
        :param style:
          font style

        :param color:
          color of text
        :param rotation:
           anticlockwise rotation
        :param justify:
          Line's justification

        :param avoid_rectangles:
          whether to avoid specified rectangles (in any case, it only works if va=top, ha=left, rotation=0)
        :param collapse_whites:
          whether multiple side-by-side withes should be considered as one
        :param paragraph_per_line:
          if true, each new line is considered a new paragraph

        :param avoid_left_of:
          tuple (x_lim, (y1, y2)). Avoid space left of x_lim between y1 and y2
        :param avoid_right_of:
          tuple (x_lim, (y1, y2)). Avoid space right of x_lim between y1 and y2

        :param va:
          Paragraph vertical alignment
        :param ha:
          Paragraph horizontal alignment
        """

        if width is None:
            width = self._width
        if justify is None:
            justify = self._justify
        if spacing is None:
            spacing = self._spacing
        if fontsize is None:
            fontsize = self._fontsize
        if color is None:
            color = self._color
        if rotation is None:
            rotation = self._rotation
        if family is None:
            family = self._family
        if weight is None:
            weight = self._weight
        if style is None:
            style = self._style
        if fontname is None:
            fontname_dict = {}
        else:
            fontname_dict = {'fontname': self._fontname}

        ax = self._axes

        def write_line(left, bottom, text_in_line, zorder=3):

            ax.text(left, bottom, text_in_line,
                    fontsize=fontsize, color=color, rotation=rotation,
                    weight=weight, style=style,
                    family=family, zorder=zorder, **fontname_dict)

        # old artists are already present in the axes and won't be moved by the posteriori vertical alignment
        old_artists = list(ax.texts)

        if ax.get_ylim()[1] < ax.get_ylim()[0] or ax.get_xlim()[1] < ax.get_xlim()[0]:
            raise NotImplementedError("paraMPL.write() is only available for plots with increasing x- and y-axis")

        # word size info
        widths, height, combined_hash = self._get_widths_height(fontsize, family, fontname,
                                                                weight, style,
                                                                words=text.split())
        space_width = widths[' ']

        # initialize position-storing object
        lp = _line_position(xy, width, height, rotation, spacing, ha, justify,
                            y_to_x_ratio=get_aspect(ax))
        lp.add_avoids(avoid_left_of, avoid_right_of)

        # add rectangles to avoid if orientation and alignment is adequate
        if va == 'top' and rotation == 0 and ha == 'left':
            if avoid_rectangles:
                lp.add_rectangles(self._rectangles)
        # if orientation is not adequate, but avoid is specified raise error
        elif avoid_left_of is not None or avoid_right_of is not None:
            raise ValueError("if using avoid areas, then va='top', ha='left', and rotation=0 are required")

        # separate and process paragraphs one at a time.
        paragraphs = split_into_paragraphs(text,
                                           collapse_whites=collapse_whites,
                                           paragraph_per_line=paragraph_per_line,
                                           )

        for paragraph in paragraphs:
            words = []
            length = 0

            # if full justified add word-by-word size and when line is completed, fill with space
            if justify == 'full':
                for word in paragraph.split(' '):
                    if length + widths[word] > lp.width_line:
                        if len(words) > 1:
                            extra_spacing = (lp.width_line - length + space_width) / (len(words) - 1)
                        else:
                            extra_spacing = 0

                        offset = 0
                        for old_word in words:
                            write_line(*lp.offset(offset=offset),
                                       old_word)
                            offset += extra_spacing + space_width + widths[old_word]

                        lp.next_line()
                        length = 0
                        words = []

                    length += widths[word] + space_width
                    words.append(word)

                write_line(*lp.offset(), ' '.join(words))
                lp.next_line()

            # if left, right, center justified then write the whole line then move it.
            else:
                for word in paragraph.split(' '):
                    if length + widths[word] > lp.width_line:
                        write_line(*lp.offset(justified_length=length - space_width),
                                   ' '.join(words))
                        lp.next_line()
                        length, words = 0, []

                    length += widths[word] + space_width
                    words.append(word)

                write_line(*lp.offset(justified_length=length - space_width),
                           ' '.join(words))
                lp.next_line()

        # get list of artists generated for these paragraphs.
        parampl_artists = [artist for artist in ax.texts if artist not in old_artists]

        # once all paragraphs are finished, do the vertical alignment
        total_height = lp.total_height()
        delta = lp.y_to_bottom_offset()

        if va == 'top':
            for artist in parampl_artists:
                artist.set_y(artist.get_position()[1] + delta - total_height)

        elif va == 'bottom':
            if delta != 0:
                for artist in parampl_artists:
                    artist.set_y(artist.get_position()[1] + delta)

        elif va == 'center':
            for artist in parampl_artists:
                artist.set_y(artist.get_position()[1] + delta - total_height / 2)

        else:
            raise ValueError(f"invalid va '{va}'. Must be 'top', 'bottom', or 'center'")

        return parampl_artists

    def _get_widths_height(self, fontsize, family, fontname,
                           weight, style,
                           words: list[str] = None,
                           ):
        text_artist = self._axes.text(0, 0, ' ',
                                      fontsize=fontsize, fontname=fontname, family=family,
                                      weight=weight, style=style,
                                      )
        combined_hash = (fontsize, family, fontname, weight, style)

        if combined_hash not in self._widths:
            text_artist.set_text(' ')
            widths: dict[str, float] = {' ': self._transformed_artist_extent(text_artist).width,
                                        '': 0,
                                        }

            text_artist.set_text('L')
            height = self._transformed_artist_extent(text_artist).height

            self._widths[combined_hash] = widths
            self._heights[combined_hash] = height
        else:
            widths = self._widths[combined_hash]

        if words is not None:
            for word in words:
                if word not in widths:
                    text_artist.set_text(word)
                    widths[word] = self._transformed_artist_extent(text_artist).width

        text_artist.remove()

        return self._widths[combined_hash], self._heights[combined_hash], combined_hash

    def _transformed_artist_extent(self, artist):
        extent = artist.get_window_extent(renderer=self._renderer)
        return extent.transformed(self._transform)
